
/* Chrysalide - Outil d'analyse de fichiers binaires
 * format.c - équivalent Python du fichier "format/format.c"
 *
 * Copyright (C) 2012-2017 Cyrille Bagard
 *
 *  This file is part of Chrysalide.
 *
 *  Chrysalide is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  Chrysalide is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#include "format.h"


#include <pygobject.h>


#include <format/format.h>


#include "executable.h"
#include "symbol.h"
#include "symiter.h"
#include "../access.h"
#include "../helpers.h"
#include "../arch/vmpa.h"



/* ---------------------------- FORMAT BINAIRE GENERIQUE ---------------------------- */


/* Enregistre une adresse comme début d'une zone de code. */
static PyObject *py_binary_format_register_code_point(PyObject *, PyObject *);

/* Ajoute un symbole à la collection du format binaire. */
static PyObject *py_binary_format_add_symbol(PyObject *, PyObject *);

/* Retire un symbole de la collection du format binaire. */
static PyObject *py_binary_format_remove_symbol(PyObject *, PyObject *);

/* Recherche le symbole correspondant à une étiquette. */
static PyObject *py_binary_format_find_symbol_by_label(PyObject *, PyObject *);

/* Recherche le symbole suivant celui lié à une adresse. */
static PyObject *py_binary_format_find_symbol_at(PyObject *, PyObject *);

/* Recherche le symbole suivant celui lié à une adresse. */
static PyObject *py_binary_format_find_next_symbol_at(PyObject *, PyObject *);

/* Recherche le symbole correspondant à une adresse. */
static PyObject *py_binary_format_resolve_symbol(PyObject *, PyObject *);

/* Indique la désignation interne du format. */
static PyObject *py_binary_format_get_name(PyObject *, void *);

/* Indique la désignation humaine du format. */
static PyObject *py_binary_format_get_description(PyObject *, void *);

/* Fournit une référence vers le contenu binaire analysé. */
static PyObject *py_binary_format_get_content(PyObject *, void *);

/* Fournit la liste de tous les symboles détectés. */
static PyObject *py_binary_format_get_symbols(PyObject *, void *);



/* ------------------ CONSERVATION DES SOUCIS DURANT LE CHARGEMENT ------------------ */


/* Etend la liste des soucis détectés avec de nouvelles infos. */
static PyObject *py_binary_format_add_error(PyObject *, PyObject *);

/* Fournit les éléments concernant tous les soucis détectés. */
static PyObject *py_binary_format_get_errors(PyObject *, void *);






/* Définit les constantes pour les types d'erreurs. */
static bool define_python_binary_format_constants(PyTypeObject *);



/* ---------------------------------------------------------------------------------- */
/*                              FORMAT BINAIRE GENERIQUE                              */
/* ---------------------------------------------------------------------------------- */


/******************************************************************************
*                                                                             *
*  Paramètres  : self = classe représentant un format.                        *
*                args = arguments fournis à l'appel.                          *
*                                                                             *
*  Description : Enregistre une adresse comme début d'une zone de code.       *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static PyObject *py_binary_format_register_code_point(PyObject *self, PyObject *args)
{
    unsigned long long pt;                  /* Adresse virtuelle du point  */
    int entry;                              /* Nature du point fourni      */
    int ret;                                /* Bilan de lecture des args.  */
    GBinFormat *format;                     /* Format de binaire manipulé  */

    ret = PyArg_ParseTuple(args, "Kp", &pt, &entry);
    if (!ret) return NULL;

    format = G_BIN_FORMAT(pygobject_get(self));

    g_binary_format_register_code_point(format, pt, entry);

    Py_RETURN_NONE;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : self = classe représentant un format.                        *
*                args = arguments fournis à l'appel.                          *
*                                                                             *
*  Description : Ajoute un symbole à la collection du format binaire.         *
*                                                                             *
*  Retour      : True si le symbole était bien localisé et a été inséré.      *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static PyObject *py_binary_format_add_symbol(PyObject *self, PyObject *args)
{
    PyObject *result;                       /* Valeur à retourner          */
    PyObject *symbol_obj;                   /* Version Python d'un symbole */
    int ret;                                /* Bilan de lecture des args.  */
    GBinFormat *format;                     /* Format de binaire manipulé  */
    GBinSymbol *symbol;                     /* Enventuel symbole trouvé    */
    bool added;                             /* Bilan de l'appel interne    */

    ret = PyArg_ParseTuple(args, "O!", get_python_binary_symbol_type(), &symbol_obj);
    if (!ret) return NULL;

    format = G_BIN_FORMAT(pygobject_get(self));
    symbol = G_BIN_SYMBOL(pygobject_get(symbol_obj));

    g_object_ref(G_OBJECT(symbol));
    added = g_binary_format_add_symbol(format, symbol);

    result = added ? Py_True : Py_False;
    Py_INCREF(result);

    return result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : self = classe représentant un format.                        *
*                args = arguments fournis à l'appel.                          *
*                                                                             *
*  Description : Retire un symbole de la collection du format binaire.        *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static PyObject *py_binary_format_remove_symbol(PyObject *self, PyObject *args)
{
    PyObject *result;                       /* Valeur à retourner          */
    PyObject *symbol_obj;                   /* Version Python d'un symbole */
    int ret;                                /* Bilan de lecture des args.  */
    GBinFormat *format;                     /* Format de binaire manipulé  */
    GBinSymbol *symbol;                     /* Enventuel symbole trouvé    */

    ret = PyArg_ParseTuple(args, "O!", get_python_binary_symbol_type(), &symbol_obj);
    if (!ret) return NULL;

    format = G_BIN_FORMAT(pygobject_get(self));
    symbol = G_BIN_SYMBOL(pygobject_get(symbol_obj));

    g_binary_format_remove_symbol(format, symbol);

    result = Py_None;
    Py_INCREF(result);

    return result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : self = classe représentant un binaire.                       *
*                args = arguments fournis à l'appel.                          *
*                                                                             *
*  Description : Recherche le symbole correspondant à une étiquette.          *
*                                                                             *
*  Retour      : Symbol trouvé si l'opération a été un succès, None sinon.    *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static PyObject *py_binary_format_find_symbol_by_label(PyObject *self, PyObject *args)
{
    PyObject *result;                       /* Valeur à retourner          */
    PyObject *label;                        /* Etiquette à retrouver       */
    int ret;                                /* Bilan de lecture des args.  */
    GBinFormat *format;                     /* Format de binaire manipulé  */
    GBinSymbol *symbol;                     /* Enventuel symbole trouvé    */
    bool found;                             /* Bilan de la recherche       */

    ret = PyArg_ParseTuple(args, "O", &label);
    if (!ret) return NULL;

    ret = PyUnicode_Check(label);
    if (!ret) return NULL;

    format = G_BIN_FORMAT(pygobject_get(self));

    found = g_binary_format_find_symbol_by_label(format, PyUnicode_DATA(label), &symbol);

    if (found)
        result = pygobject_new(G_OBJECT(symbol));
    else
    {
        result = Py_None;
        Py_INCREF(result);
    }

    return result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : self = classe représentant un binaire.                       *
*                args = arguments fournis à l'appel.                          *
*                                                                             *
*  Description : Recherche le symbole suivant celui lié à une adresse.        *
*                                                                             *
*  Retour      : Symbol trouvé si l'opération a été un succès, None sinon.    *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static PyObject *py_binary_format_find_symbol_at(PyObject *self, PyObject *args)
{
    PyObject *result;                       /* Valeur à retourner          */
    GBinFormat *format;                     /* Format de binaire manipulé  */
    exe_cv_info_t conv;                     /* Informations de conversion  */
    int ret;                                /* Bilan de lecture des args.  */
    GBinSymbol *symbol;                     /* Enventuel symbole trouvé    */
    bool found;                             /* Bilan de la recherche       */

    format = G_BIN_FORMAT(pygobject_get(self));

    conv.format = G_IS_EXE_FORMAT(format) ? G_EXE_FORMAT(format) : NULL;

    ret = PyArg_ParseTuple(args, "O&", convert_to_vmpa_using_executable, &conv);
    if (!ret) return NULL;

    found = g_binary_format_find_symbol_at(format, conv.vmpa, &symbol);

    if (found)
    {
        result = pygobject_new(G_OBJECT(symbol));
        g_object_unref(G_OBJECT(symbol));
    }
    else
    {
        result = Py_None;
        Py_INCREF(result);
    }

    return result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : self = classe représentant un binaire.                       *
*                args = arguments fournis à l'appel.                          *
*                                                                             *
*  Description : Recherche le symbole suivant celui lié à une adresse.        *
*                                                                             *
*  Retour      : Symbol trouvé si l'opération a été un succès, None sinon.    *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static PyObject *py_binary_format_find_next_symbol_at(PyObject *self, PyObject *args)
{
    PyObject *result;                       /* Valeur à retourner          */
    GBinFormat *format;                     /* Format de binaire manipulé  */
    exe_cv_info_t conv;                     /* Informations de conversion  */
    int ret;                                /* Bilan de lecture des args.  */
    GBinSymbol *symbol;                     /* Enventuel symbole trouvé    */
    bool found;                             /* Bilan de la recherche       */

    format = G_BIN_FORMAT(pygobject_get(self));

    conv.format = G_IS_EXE_FORMAT(format) ? G_EXE_FORMAT(format) : NULL;

    ret = PyArg_ParseTuple(args, "O&", convert_to_vmpa_using_executable, &conv);
    if (!ret) return NULL;

    found = g_binary_format_find_next_symbol_at(format, conv.vmpa, &symbol);

    if (found)
    {
        result = pygobject_new(G_OBJECT(symbol));
        g_object_unref(G_OBJECT(symbol));
    }
    else
    {
        result = Py_None;
        Py_INCREF(result);
    }

    return result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : self = classe représentant un format binaire.                *
*                args = arguments fournis à l'appel.                          *
*                                                                             *
*  Description : Recherche le symbole correspondant à une adresse.            *
*                                                                             *
*  Retour      : Tuple (nom, décalage) ou Py_None.                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static PyObject *py_binary_format_resolve_symbol(PyObject *self, PyObject *args)
{
    PyObject *result;                       /* Valeur à retourner          */
    GBinFormat *format;                     /* Format de binaire manipulé  */
    exe_cv_info_t conv;                     /* Informations de conversion  */
    int strict;                             /* Tolérance acceptée          */
    int ret;                                /* Bilan de lecture des args.  */
    GBinSymbol *symbol;                     /* Enventuel symbole trouvé    */
    phys_t diff;                            /* Décalage éventuel mesuré    */
    bool found;                             /* Bilan de la recherche       */

    format = G_BIN_FORMAT(pygobject_get(self));

    conv.format = G_IS_EXE_FORMAT(format) ? G_EXE_FORMAT(format) : NULL;

    ret = PyArg_ParseTuple(args, "O&p", convert_to_vmpa_using_executable, &conv, &strict);
    if (!ret) return NULL;

    found = g_binary_format_resolve_symbol(format, conv.vmpa, strict, &symbol, &diff);

    if (found)
    {
        result = PyTuple_New(2);
        PyTuple_SetItem(result, 0, pygobject_new(G_OBJECT(symbol)));
        PyTuple_SetItem(result, 1, PyLong_FromUnsignedLongLong(diff));

        g_object_unref(G_OBJECT(symbol));

    }
    else
    {
        result = Py_None;
        Py_INCREF(result);
    }

    return result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : self    = objet Python concerné par l'appel.                 *
*                closure = non utilisé ici.                                   *
*                                                                             *
*  Description : Indique la désignation interne du format.                    *
*                                                                             *
*  Retour      : Description du format.                                       *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static PyObject *py_binary_format_get_name(PyObject *self, void *closure)
{
    PyObject *result;                       /* Trouvailles à retourner     */
    GBinFormat *format;                     /* Format de binaire manipulé  */
    const char *name;                       /* Description interne         */

    format = G_BIN_FORMAT(pygobject_get(self));

    name = g_binary_format_get_name(format);

    result = PyUnicode_FromString(name);

    return result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : self    = objet Python concerné par l'appel.                 *
*                closure = non utilisé ici.                                   *
*                                                                             *
*  Description : Indique la désignation humaine du format.                    *
*                                                                             *
*  Retour      : Description du format.                                       *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static PyObject *py_binary_format_get_description(PyObject *self, void *closure)
{
    PyObject *result;                       /* Trouvailles à retourner     */
    GBinFormat *format;                     /* Format de binaire manipulé  */
    const char *desc;                       /* Description humaine         */

    format = G_BIN_FORMAT(pygobject_get(self));

    desc = g_binary_format_get_description(format);

    result = PyUnicode_FromString(desc);

    return result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : self    = objet Python concerné par l'appel.                 *
*                closure = non utilisé ici.                                   *
*                                                                             *
*  Description : Fournit une référence vers le contenu binaire analysé.       *
*                                                                             *
*  Retour      : Gestionnaire de contenu binaire en place.                    *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static PyObject *py_binary_format_get_content(PyObject *self, void *closure)
{
    PyObject *result;                       /* Trouvailles à retourner     */
    GBinFormat *format;                     /* Format de binaire manipulé  */
    GBinContent *content;                   /* Instance GLib correspondante*/

    format = G_BIN_FORMAT(pygobject_get(self));

    content = g_binary_format_get_content(format);

    result = pygobject_new(G_OBJECT(content));

    g_object_unref(content);

    return result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : self    = classe représentant un format binaire.             *
*                closure = adresse non utilisée ici.                          *
*                                                                             *
*  Description : Fournit la liste de tous les symboles détectés.              *
*                                                                             *
*  Retour      : Tableau créé ou NULL si aucun symbole trouvé.                *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static PyObject *py_binary_format_get_symbols(PyObject *self, void *closure)
{
    PyObject *result;                       /* Instance Python à retourner */
    PyTypeObject *iterator_type;            /* Type Python de l'itérateur  */
    PyObject *args;                         /* Liste des arguments d'appel */

    iterator_type = get_python_sym_iterator_type();

    args = Py_BuildValue("On", self, 0);

    result = PyObject_CallObject((PyObject *)iterator_type, args);

    Py_DECREF(args);

    return result;

}



/* ---------------------------------------------------------------------------------- */
/*                    CONSERVATION DES SOUCIS DURANT LE CHARGEMENT                    */
/* ---------------------------------------------------------------------------------- */


/******************************************************************************
*                                                                             *
*  Paramètres  : self = architecture concernée par la procédure.              *
*                args = instruction représentant le point de départ.          *
*                                                                             *
*  Description : Etend la liste des soucis détectés avec de nouvelles infos.  *
*                                                                             *
*  Retour      : None.                                                        *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static PyObject *py_binary_format_add_error(PyObject *self, PyObject *args)
{
    BinaryFormatError type;                 /* Type d'erreur détectée      */
    vmpa2t addr;                            /* Position d'une erreur       */
    const char *desc;                       /* Description d'une erreur    */
    int ret;                                /* Bilan de lecture des args.  */
    GBinFormat *format;                     /* Format binaire manipulé     */

    ret = PyArg_ParseTuple(args, "IO&s", &type, convert_any_to_vmpa, &addr, &desc);
    if (!ret) return NULL;

    format = G_BIN_FORMAT(pygobject_get(self));

    g_binary_format_add_error(format, type, &addr, desc);

    Py_RETURN_NONE;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : self    = objet Python concerné par l'appel.                 *
*                closure = non utilisé ici.                                   *
*                                                                             *
*  Description : Fournit les éléments concernant tous les soucis détectés.    *
*                                                                             *
*  Retour      : Liste des erreurs relevées au niveau de l'assembleur.        *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static PyObject *py_binary_format_get_errors(PyObject *self, void *closure)
{
    PyObject *result;                       /* Instance Python à retourner */
    GBinFormat *format;                     /* Format binaire manipulé     */
    size_t count;                           /* Nombre d'éléments à traiter */
    size_t i;                               /* Boucle de parcours          */
#ifndef NDEBUG
    bool status;                            /* Bilan d'un appel            */
#endif
    BinaryFormatError type;                 /* Type d'erreur détectée      */
    vmpa2t addr;                            /* Position d'une erreur       */
    char *desc;                             /* Description d'une erreur    */
    PyObject *error;                        /* Nouvelle erreur à rajouter  */

    format = G_BIN_FORMAT(pygobject_get(self));

    g_binary_format_lock_errors(format);

    count = g_binary_format_count_errors(format);

    result = PyTuple_New(count);

    for (i = 0; i < count; i++)
    {
#ifndef NDEBUG
        status = g_binary_format_get_error(format, i, &type, &addr, &desc);
        assert(status);
#else
        g_binary_format_get_error(format, i, &type, &addr, &desc);
#endif

        error = Py_BuildValue("IO&s", type, build_from_internal_vmpa, &addr, desc);

        PyTuple_SetItem(result, i, error);

    }

    g_binary_format_unlock_errors(format);

    return result;

}







/******************************************************************************
*                                                                             *
*  Paramètres  : obj_type = type dont le dictionnaire est à compléter.        *
*                                                                             *
*  Description : Définit les constantes pour les types d'erreurs.             *
*                                                                             *
*  Retour      : true en cas de succès de l'opération, false sinon.           *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static bool define_python_binary_format_constants(PyTypeObject *obj_type)
{
    bool result;                            /* Bilan à retourner           */

    result = true;

    result &= PyDict_AddULongMacro(obj_type, BFE_SPECIFICATION);
    result &= PyDict_AddULongMacro(obj_type, BFE_STRUCTURE);

    return result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : -                                                            *
*                                                                             *
*  Description : Fournit un accès à une définition de type à diffuser.        *
*                                                                             *
*  Retour      : Définition d'objet pour Python.                              *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

PyTypeObject *get_python_binary_format_type(void)
{
    static PyMethodDef py_bin_format_methods[] = {
        {
            "register_code_point", py_binary_format_register_code_point,
            METH_VARARGS,
            "register_code_point($self, pt, entry, /)\n--\n\nRegister a virtual address as entry point or basic point."
        },
        {
            "add_symbol", py_binary_format_add_symbol,
            METH_VARARGS,
            "add_symbol($self, symbol, /)\n--\n\nRegister a new symbol for the format."
        },
        {
            "remove_symbol", py_binary_format_remove_symbol,
            METH_VARARGS,
            "remove_symbol($self, symbol, /)\n--\n\nUnregister a symbol from the format."
        },
        {
            "find_symbol_by_label", py_binary_format_find_symbol_by_label,
            METH_VARARGS,
            "find_symbol_by_label($self, label, /)\n--\n\nFind a symbol by its label."
        },
        {
            "find_symbol_at", py_binary_format_find_symbol_at,
            METH_VARARGS,
            "find_symbol_at($self, addr, /)\n--\n\nFind a symbol at a given address."
        },
        {
            "find_next_symbol_at", py_binary_format_find_next_symbol_at,
            METH_VARARGS,
            "find_next_symbol_at($self, addr, /)\n--\n\nFind the symbol next to the one found at a given address."
        },
        {
            "resolve_symbol", py_binary_format_resolve_symbol,
            METH_VARARGS,
            "resolve_symbol($self, addr, strict, /)\n--\n\nSearch a position inside a routine by a given address."
        },
        {
            "add_error", py_binary_format_add_error,
            METH_VARARGS,
            "add_error($self, type, addr, desc, /)\n--\n\nExtend the list of detected disassembling errors."
        },
        { NULL }
    };

    static PyGetSetDef py_bin_format_getseters[] = {
        {
            "name", py_binary_format_get_name, NULL,
            "Internal name of the binary format.", NULL
        },
        {
            "description", py_binary_format_get_description, NULL,
            "Human description of the binary format.", NULL
        },
        {
            "content", py_binary_format_get_content, NULL,
            "Content of the binary format.", NULL
        },
        {
            "symbols", py_binary_format_get_symbols, NULL,
            "Iterable list of all symbols found in the binary format.", NULL
        },
        {
            "errors", py_binary_format_get_errors, NULL,
            "List of all detected errors which occurred while loading the binary.", NULL
        },
        { NULL }
    };

    static PyTypeObject py_bin_format_type = {

        PyVarObject_HEAD_INIT(NULL, 0)

        .tp_name        = "pychrysalide.format.BinFormat",
        .tp_basicsize   = sizeof(PyGObject),

        .tp_flags       = Py_TPFLAGS_DEFAULT | Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_BASETYPE,

        .tp_doc         = "PyChrysalide binary format",

        .tp_methods     = py_bin_format_methods,
        .tp_getset      = py_bin_format_getseters

    };

    return &py_bin_format_type;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : module = module dont la définition est à compléter.          *
*                                                                             *
*  Description : Prend en charge l'objet 'pychrysalide.format.BinFormat'.     *
*                                                                             *
*  Retour      : Bilan de l'opération.                                        *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

bool ensure_python_binary_format_is_registered(void)
{
    PyTypeObject *type;                     /* Type Python 'BinFormat'     */
    PyObject *module;                       /* Module à recompléter        */
    PyObject *dict;                         /* Dictionnaire du module      */

    type = get_python_binary_format_type();

    if (!PyType_HasFeature(type, Py_TPFLAGS_READY))
    {
        module = get_access_to_python_module("pychrysalide.format");

        APPLY_ABSTRACT_FLAG(type);

        dict = PyModule_GetDict(module);

        if (!register_class_for_pygobject(dict, G_TYPE_BIN_FORMAT, type, &PyGObject_Type))
            return false;

        if (!define_python_binary_format_constants(type))
            return false;

    }

    return true;

}
