
/* Chrysalide - Outil d'analyse de fichiers binaires
 * panels.c - gestion d'ensemble de tous les panneaux pour l'éditeur
 *
 * Copyright (C) 2016-2017 Cyrille Bagard
 *
 *  This file is part of Chrysalide.
 *
 *  Chrysalide is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  Chrysalide is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#include "panels.h"


#include "global.h"
#include "items.h"
#include "../panels/bintree.h"
#include "../panels/bookmarks.h"
#include "../panels/errors.h"
#include "../panels/glance.h"
#include "../panels/history.h"
#include "../panels/log.h"
#include "../panels/panel-int.h"
#include "../panels/regedit.h"
#include "../panels/strings.h"
#include "../panels/symbols.h"
#include "../panels/welcome.h"
#include "../../core/params.h"
#include "../../gtkext/gtkdockable.h"



/* Liste des panneaux en place. */
static GPanelItem *_panels_list = NULL;



/******************************************************************************
*                                                                             *
*  Paramètres  : -                                                            *
*                                                                             *
*  Description : Charge les principaux panneaux de l'éditeur.                 *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

void load_main_panels(void)
{
    GGenConfig *config;                     /* Configuration globale       */
    GPanelItem *item;                       /* Panneau de base à charger   */

    config = get_main_configuration();

    item = g_welcome_panel_new();
    register_panel_item(item, config);

    item = g_log_panel_new();
    register_panel_item(item, config);

    item = g_regedit_panel_new();
    register_panel_item(item, config);

    item = g_symbols_panel_new();
    register_panel_item(item, config);

    item = g_history_panel_new();
    register_panel_item(item, config);

    item = g_strings_panel_new();
    register_panel_item(item, config);

    item = g_glance_panel_new();
    register_panel_item(item, config);

    item = g_bookmarks_panel_new();
    register_panel_item(item, config);

    item = g_bintree_panel_new();
    register_panel_item(item, config);

    item = g_error_panel_new();
    register_panel_item(item, config);

}


/******************************************************************************
*                                                                             *
*  Paramètres  : item   = composant à présenter à l'affichage.                *
*                config = configuration à compléter.                          *
*                                                                             *
*  Description : Enregistre un panneau comme partie intégrante de l'éditeur.  *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

void register_panel_item(GPanelItem *item, GGenConfig *config)
{
    GEditorItem *parent;                    /* Autre version de l'élément  */

    parent = G_EDITOR_ITEM(item);

    /* Enregistre correctement le tout */
    register_editor_item(parent);
    panels_list_add_tail(item, &_panels_list);

    activate_panel_item(item, config);

}


/******************************************************************************
*                                                                             *
*  Paramètres  : item   = composant à présenter à l'affichage.                *
*                config = configuration à compléter.                          *
*                                                                             *
*  Description : Intègre un panneau comme partie intégrante de l'éditeur.     *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

void activate_panel_item(GPanelItem *item, GGenConfig *config)
{
    GtkTiledGrid *grid;                     /* Composant d'affichage       */

    grid = get_tiled_grid();

    g_signal_connect_swapped(item, "dock-request", G_CALLBACK(gtk_tiled_grid_add), grid);
    g_signal_connect_swapped(item, "undock-request", G_CALLBACK(gtk_tiled_grid_remove), grid);

    gtk_dockable_setup_dnd(GTK_DOCKABLE(item));

    gtk_panel_item_setup_configuration(item, config);

}


/******************************************************************************
*                                                                             *
*  Paramètres  : skip   = saute le panneau d'accueil lors du parcours ?       *
*                handle = routine à appeler pour chaque panneau.              *
*                data   = données fournies pour accompagner cet appel.        *
*                                                                             *
*  Description : Effectue le parcours de tous les panneaux chargés.           *
*                                                                             *
*  Retour      : true si le parcours a été total, false sinon.                *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

bool _browse_all_item_panels(bool skip, handle_panel_item_fc handle, void *data)
{
    bool result;                            /* Résultat à renvoyer         */
    GPanelItem *welcome;                    /* Panneau d'accueil           */
    GPanelItem *iter;                       /* Boucle de parcours          */

    result = true;

    if (skip)
        welcome = get_panel_item_by_name(PANEL_WELCOME_ID);
    else
        welcome = NULL;

    panels_list_for_each(iter, _panels_list)
    {
        if (skip && iter == welcome)
            continue;

        result = handle(iter, data);

        if (!result) break;

    }

    g_clear_object(&welcome);

    return result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : name = désignation courte servant de clef.                   *
*                                                                             *
*  Description : Recherche un panneau à partir de son nom court.              *
*                                                                             *
*  Retour      : Panneau trouvé ou NULL si aucun.                             *
*                                                                             *
*  Remarques   : Le parcours peut se faire aussi depuis la classe parente,    *
*                mais il est plus rapide par ici.                             *
*                                                                             *
******************************************************************************/

GPanelItem *get_panel_item_by_name(const char *name)
{
    GPanelItem *result;                     /* Trouvaille à retourner      */

    bool look_for_named_panel(GPanelItem *item, GPanelItem **found)
    {
        const char *key;                    /* Clef à utiliser             */
        bool status;                        /* Bilan de la comparaison     */

        key = g_editor_item_get_name(G_EDITOR_ITEM(item));

        if (strcmp(key, name) == 0)
        {
            *found = item;
            status = false;
        }
        else
            status = true;

        return status;

    }

    result = NULL;

    browse_all_item_panels((handle_panel_item_fc)look_for_named_panel, &result);

    if (result != NULL)
        g_object_ref(G_OBJECT(result));

    return result;

}
